% Learn sparse Ising model with stochastic approximate D(logZ)
% Exact sampling:
%   sampling.method = 'exact'
% Importance sampling by uniform distribution:
%   sampling.method = 'uniform'
%   sampling.samples
% Importance sampling by with mean-field distribution:
%   sampling.method = 'meanfield'
%   sampling.samples
% Gibbs sampler by starting with mean-field distribution:
%   sampling.method = 'gibbs'
%   sampling.samples
%   sampling.iterations
% All methods
%   sampling.grad_error = true, to measure gradient error
%   sampling.obj_fun = true, to measure obj_fun
%   sampling.truth_P = ground truth probability
%
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function [W b rule addtl] = learn_sproxgrad_ising(X,rho,Touter,sampling,epsilon,accelerated,verbose);

if ~issign(X)
    error('Only +1 and -1 are accepted in dataset');
end
if ~isfield(sampling,'grad_error')
    sampling.grad_error = false;
end
if ~isfield(sampling,'obj_fun')
    sampling.obj_fun = false;
end
if ~exist('accelerated','var')
    accelerated = false;
end
if ~exist('verbose','var')
    verbose = false;
end
[T N] = size(X);
S = X.'*X/T - eye(N);
mu = reshape(sum(X,1)/T,[],1);
W = zeros(N);
b = zeros(N,1);
sum_W = zeros(N);
sum_b = zeros(N,1);
epsilon_G = epsilon/(max(2,rho)*N);
alpha_t = epsilon_G;
if strcmp(sampling.method,'uniform')
    P1 = ones(N,1)/2;
end
addtl.grad_error = [];
addtl.nll = [];
addtl.obj_fun = [];
addtl.kld = [];
W_prev = W;
b_prev = b;
for t = 1:Touter
    if verbose
        if mod(t,10) == 0
            fprintf('.');
        end
    end
    if isfield(sampling,'samples')
        if isnumeric(sampling.samples)
            sampling_samples = sampling.samples;
        else
            sampling_samples = ceil(eval(sampling.samples));
        end
    end
    if accelerated
        Wy = W + (t-1)/(t+2)*(W-W_prev);
        by = b + (t-1)/(t+2)*(b-b_prev);
    else
        Wy = W;
        by = b;
    end
    while true
        if strcmp(sampling.method,'exact')
            [tmp DW Db] = log_partition_ising(Wy,by);
        elseif strcmp(sampling.method,'uniform')
            [DW Db] = log_partition_importsam_ising(Wy,by,sampling_samples,P1);
        elseif strcmp(sampling.method,'meanfield')
            [tmp tmp m] = log_partition_meanfield_ising(Wy,by);
            P1 = (1+m)/2;
            [DW Db] = log_partition_importsam_ising(Wy,by,sampling_samples,P1);
        elseif strcmp(sampling.method,'gibbs')
            [tmp tmp m] = log_partition_meanfield_ising(Wy,by);
            P1 = (1+m)/2;
            [DW Db] = log_partition_gibbssam_ising(Wy,by,sampling_samples,P1,sampling.iterations);
        else
            error('Unknown sampling');
        end
        if all(~isnan(DW(:))) & all(~isnan(Db))
            break
        end
        fprintf('SAMPLING returned NaNs, TRYING AGAIN...\n');
    end
    if sampling.grad_error
        [tmp DW_truth Db_truth] = log_partition_ising(Wy,by);
        addtl.grad_error(t,1) = sqrt(sum(DW(:)-DW_truth(:)).^2);
        addtl.grad_error(t,2) = sqrt(sum(Db(:)-Db_truth(:)).^2);
    end
    alpha_tnext = epsilon_G; %/sqrt(t+1);
    W_prev = W;
    b_prev = b;
    W = Wy - alpha_t*(DW-S/2);
    W = sign(W).*max(0,abs(W) - rho*alpha_tnext);
    b = by - alpha_t*(Db-mu);
    sum_W = sum_W + W;
    sum_b = sum_b + b;
    alpha_t = alpha_tnext;
    if sampling.obj_fun
        min_prob = 1e-12;
        Wo = W;
        bo = b;
        [logZ Q] = log_partition_ising(Wo,bo);
        if min(Q) < min_prob
            Q = Q - min(Q) + min_prob;
            Q = Q / sum(Q);
        end
        n1 = logZ - sum(sum(S.*Wo))/2 - sum(mu.*bo);
        o1 = n1 + rho*sum(abs(Wo(:)));
        k1 = sum(sampling.truth_P.*(log(sampling.truth_P)-log(Q)));
        Wo = sum_W / t;
        bo = sum_b / t;
        [logZ Q] = log_partition_ising(Wo,bo);
        if min(Q) < min_prob
            Q = Q - min(Q) + min_prob;
            Q = Q / sum(Q);
        end
        n2 = logZ - sum(sum(S.*Wo))/2 - sum(mu.*bo);
        o2 = n2 + rho*sum(abs(Wo(:)));
        k2 = sum(sampling.truth_P.*(log(sampling.truth_P)-log(Q)));
        addtl.nll = [addtl.nll; n1 n2];
        addtl.obj_fun = [addtl.obj_fun; o1 o2];
        addtl.kld = [addtl.kld; k1 k2];
    end
end
if verbose
    fprintf('\n');
end
last_W = W;
last_b = b;
clear W b;
W{1} = last_W;
b{1} = last_b;
W{2} = sum_W / Touter;
b{2} = sum_b / Touter;
rule = {'last','avg'};
